#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220820
__all__ = ['dimensionless', 'percentage']

from ..dictionaries import dictionary
from ..unit_class import unit
from ..errors import WrongUnitsError


class dimensionless(unit):
    classUnits = dictionary['dimensionless']
    def __init__(self, value, units=None):
        self.name = 'dimensionless'
        self.kind = dimensionless
        self.value = self.checkValue(value)
        if units is None:
            units = 'dimensionless'
        self.unit = self.checkUnit(units)

    def convert(self, newunit=None):
        if newunit is not None and type(newunit) is not str:
            try:
                newunit = newunit.unit
            except:
                raise WrongUnitsError("'" + str(newunit) + "' for '" + str(self.name) + "'")
        if newunit is None or len(newunit) == 0:
            return self.value
        elif newunit in dictionary['percentage']:
            return percentage(self.value * 100, newunit)
        elif newunit in dictionary['dimensionless']:
            return dimensionless(self.value, newunit)
        else:
            from .define import units
            return units(self.value, newunit)

    def to(self, newunit=None):
        return self.convert(newunit)


class percentage(dimensionless):
    classUnits = dictionary['percentage']
    def __init__(self, value, units=None):
        self.name = 'percentage'
        self.kind = percentage
        self.value = self.checkValue(value) / 100
        if units is None:
            units = 'percentage'
        self.unit = self.checkUnit(units)

    def __repr__(self):
        return str(self.value * 100) + '_' + str(self.unit)

    def __str__(self) :
        return str(self.value * 100) + '_' + str(self.unit)

    def __neg__(self):
        return self.kind(self.value * -100, self.unit)

    def __abs__(self):
        return self.kind(abs(self.value) * 100, self.unit)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['time']

from ..dictionaries import dictionary
from ..unit_class import unit


class time(unit):
    classUnits = dictionary['time']
    def __init__(self, value, units):
        self.name = 'time'
        self.kind = time
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 15:57:27 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.2'
__release__ = 20220924
__all__ = ['converter', 'convertible', 'convertUnit']


from .database import unitsNetwork
from .dictionaries import dictionary, temperatureRatioConversions
from .searches import BFS, printPath
from .errors import NoConversionFoundError
import numpy as np
from functools import reduce


def _cleanPrintConversionPath(printConversionPath):
    if printConversionPath is None:
        printConversionPath = unitsNetwork.print
    elif type(printConversionPath) is not bool:
        if type(printConversionPath) in [int, float]:
            if printConversionPath == 1:
                printConversionPath = True
            else:
                printConversionPath = False
        else:
            printConversionPath = True
    return printConversionPath


def convertible(fromUnit, toUnit, PrintPath=False):
    try:
        converter(1, fromUnit, toUnit, False)
        return True
    except:
        return False


def convertUnit(value, fromUnit, toUnit, printConversionPath=False):
    """
    returns the received value (integer, float, array, Series, DataFrame, etc)
    transformed from the units 'fromUnit' to the units 'toUnits.

    Parameters
    ----------
    value : int, float, array, Series, DataFrame, etc
        the value to be converted.
    fromUnit : str
        the units of the provided value.
    toUnit : str
        the units to convert the value.
    printConversionPath : bool, optional
        Set to True to show the path used for conversion. The default is False.

    Returns
    -------
    converted_value : int, float, array, Series, DataFrame ...
        the converted value

    """
    printConversionPath = _cleanPrintConversionPath(printConversionPath)

    try:
        conv = converter(value, fromUnit, toUnit, printConversionPath)
        if conv is not None:
            return conv
    except NoConversionFoundError:
        return None


def convertUnit_for_SimPandas(value, fromUnit, toUnit, printConversionPath=False):
    printConversionPath = _cleanPrintConversionPath(printConversionPath)

    try:
        conv = converter(value, fromUnit, toUnit, printConversionPath)
        if conv is not None:
            return conv
        else:
            return value
    except NoConversionFoundError:
        return value


def _applyConversion(value, conversionPath, printConversionPath=False):
    if printConversionPath:
        print( "\n converting from '" + str(conversionPath[0]) + "' to '" + str(conversionPath[-1]) + "'\n  " + printPath(conversionPath) )
    for conversion in range(len(conversionPath) - 1):
        value = unitsNetwork.convert(value, conversionPath[conversion], conversionPath[conversion + 1])
    return value


def _lambdaConversion(conversionPath, printConversionPath=False):
    if printConversionPath:
        print("\n converting from '" + str(conversionPath[0]) + "' to '" + str(conversionPath[-1]) + "'\n  " + printPath(conversionPath))
    bigLambda = []
    for i in range(len(conversionPath) - 1):
        bigLambda += [unitsNetwork.conversion(conversionPath[i],conversionPath[i + 1])]
    return lambda x: _lambdaLoop(x, bigLambda[:])


def _lambdaLoop(x, LambdaList):
    for L in LambdaList:
        x=L(x)
    return x


def _splitRatio(unit):
    unitUp , unitDown = unit.split('/')
    return unitUp.strip(), unitDown.strip()


def _splitProduct(unit):
    unitA , unitB = unit.split('*')
    return unitA.strip(), unitB.strip()


def _splitUnit(unit):
    result = None
    for c in unit:
        if c in '*/':
            result.append(c)
        elif result is None:
            result = [c]
        elif result[-1] in '*/':
            result.append(c)
        else:
            result[-1] = result[-1] + c
    return result


def _getPairChild(unit):

    # get the unit node if the name received
    unit = unitsNetwork.getNode(unit) if type(unit) is str else unit

    # get pair of units children
    pairChild = list(filter(lambda u: '/' in u or '*' in u, [u.getName() for u in unitsNetwork.childrenOf(unit)]))

    ## if pair of units child is found, return the one with the shorter name
    if len(pairChild) > 0:
        pairChild = sorted(pairChild, key=len)[0]

    ## if no children found at this level, look for children in next level
    else:
        for child in unitsNetwork.childrenOf(unit):
            pairGrandchild = list(filter(lambda u: '/' in u or '*' in u, [u.getName() for u in unitsNetwork.childrenOf(child)]))
            if len(pairGrandchild) > 0:
                pairChild = sorted(pairGrandchild, key=len)[0]
                break

    if type(pairChild) is str:
        return pairChild


def _get_conversion(value, fromUnit, toUnit, printConversionPath=None, getPath=False):
    # specific cases for quick conversions
    ## no conversion required if 'from' and 'to' units are the same units
    if fromUnit == toUnit:
        return (lambda x: x) if value is None else value

    ## no conversion required if 'from' and 'to' units are dates
    if fromUnit in dictionary['date'] and toUnit in dictionary['date']:
        return (lambda x: x) if value is None else value

    ## special case for temperature ratios
    if '/' in fromUnit and len(fromUnit.split('/'))==2 and fromUnit.split('/')[0] in dictionary['temperature'] \
        and '/' in toUnit and len(toUnit.split('/'))==2 and toUnit.split('/')[0] in dictionary['temperature']:
            t1, d1 = fromUnit.split('/')
            t2, d2 = toUnit.split('/')
            num = temperatureRatioConversions[(t1, t2)]
            den = _get_conversion(1, d1, d2, printConversionPath=printConversionPath, getPath=getPath)
            if num is None or den is None:
                return None  # raise NoConversionFound("from '" + str(d1) + "' to '" + str(d2) + "'")
            return (lambda x: x * num / den) if value is None else (value * num / den)

    ## from dimensionless/None to some units or viceversa (to allow assign units to dimensionless numbers)
    if fromUnit is None or toUnit is None:
        return (lambda x: x) if value is None else value
    if (fromUnit.lower() in dictionary['dimensionless'] or toUnit.lower() in dictionary['dimensionless']) and (fromUnit is None or toUnit is None):
        return (lambda x: x) if value is None else value

    ## from dimensionless to percentage or viceversa
    if (fromUnit is None or fromUnit.lower() in dictionary['dimensionless']) and toUnit.lower() in dictionary['percentage']:
        return (lambda x: x * 100) if value is None else value * 100
    if fromUnit.lower() in dictionary['percentage'] and (toUnit is None or toUnit.lower() in dictionary['dimensionless']):
        return (lambda x: x / 100) if value is None else value / 100

    ## from dimensionless to ratio of same units
    if fromUnit.lower() in dictionary['dimensionless'] and '/' in toUnit and len(toUnit.split('/')) == 2 and toUnit.lower().split('/')[0].strip(' ()') == toUnit.lower().split('/')[1].strip(' ()'):
        return (lambda x: x) if value is None else value

    ## from ratio of same units to dimensionless
    if toUnit.lower() in dictionary['dimensionless'] and '/' in fromUnit and len(fromUnit.split('/')) == 2 and fromUnit.lower().split('/')[0].strip(' ()') == fromUnit.lower().split('/')[1].strip(' ()'):
        return (lambda x: x) if value is None else value

    ## if inverted ratios
    if ('/' in fromUnit and len(fromUnit.split('/')) == 2) and ('/' in toUnit and len(toUnit.split('/')) == 2) and (fromUnit.split('/')[0].strip() == toUnit.split('/')[1].strip()) and (fromUnit.split('/')[1].strip() == toUnit.split('/')[0].strip()):
        return (lambda x: 1/x) if value is None else 1/value

    # check if already solved and memorized
    if not getPath and (fromUnit, toUnit) in unitsNetwork.Memory:
        conversionLambda = unitsNetwork.Memory[(fromUnit,toUnit)]
        return conversionLambda if value is None else conversionLambda(value)

    # check if path is alredy defined in network
    if unitsNetwork.hasNode(fromUnit) and unitsNetwork.hasNode(toUnit):
        conversionPath = BFS(unitsNetwork, unitsNetwork.getNode(fromUnit), unitsNetwork.getNode(toUnit))
    else:
        conversionPath = None

    ## return conversion if found in network
    if conversionPath is not None:
        unitsNetwork.Memory[(fromUnit, toUnit)] = _lambdaConversion(conversionPath, printConversionPath=False)
        if getPath:
            return conversionPath
        return _applyConversion(value, conversionPath, printConversionPath) if value is not None else unitsNetwork.Memory[(fromUnit,toUnit)]


def converter(value, fromUnit, toUnit, printConversionPath=None):
    """
    returns the received value (integer, float, array, series, dataframe, etc)
    transformed from the units 'fromUnit' to the units 'toUnits.
    """
    if hasattr(value, '__iter__') and type(value) is not np.array:
        value = np.array(value)

    printConversionPath = _cleanPrintConversionPath(printConversionPath)

    # cleaning inputs
    ## strip off the parentesis, the string o
    if type(fromUnit) is str and fromUnit not in ('"',"'"):
        fromUnit = fromUnit.strip("( ')").strip('( ")').strip("'")
    else:
        fromUnit = fromUnit.strip("( )")
    if type(toUnit) is str and toUnit not in ('"',"'"):
        toUnit = toUnit.strip("( ')").strip('( ")').strip("'")
    else:
        toUnit = toUnit.strip("( )")

    # reset memory for this variable
    unitsNetwork.previous = []

    # try to convert
    conversion = _get_conversion(value, fromUnit, toUnit,
                                 printConversionPath=printConversionPath)

    ## if convertion found
    if conversion is not None:
        return conversion

    # else, if conversion path not found, start a new search part by part
    ## check if pair from-to already visited
    if (fromUnit, toUnit) in unitsNetwork.previous:
        AllowRecursion = 0  # stop recursion here
    ## append this pair to this search history
    unitsNetwork.previous.append((fromUnit, toUnit))

    listConversion = []
    splitFrom = _splitUnit(fromUnit)
    splitTo = _splitUnit(toUnit)
    usedTo = []
    failed = False
    for f in range(len(splitFrom)):
        flag = False
        if splitFrom[f] in '*/':
            continue
        for t in range(len(splitTo)):
            if t in usedTo:
                continue
            if splitTo[t] in '*/':
                continue
            conversion = _get_conversion(1, splitFrom[f], splitTo[t],
                                         printConversionPath=printConversionPath,
                                         getPath=False)
            if conversion is not None:
                flag = True
                if (f > 0 and splitFrom[f-1] == '/') or (t > 0 and splitTo[t-1] == '/'):
                    conversion = 1 / conversion
                listConversion.append(conversion)
                usedTo.append(t)
                break
        if not flag:
            failed = True
            break

    if listConversion != [] and not failed:
        conversionFactor = reduce(lambda x, y : x * y, listConversion)
        unitsNetwork.Memory[(fromUnit, toUnit)] = lambda x : x * conversionFactor
        return value * conversionFactor if value is not None else unitsNetwork.Memory[(fromUnit, toUnit)]

    # look for one to pair conversion
    if ('/' in toUnit or '*' in toUnit) and ('/' not in fromUnit and '*' not in fromUnit):
        fromUnitChild = _getPairChild(fromUnit)
        if fromUnitChild is not None:
            baseConversion = _get_conversion(1, fromUnit, fromUnitChild)
            sep = '/' if '/' in fromUnitChild else '*'
            fromNum, fromDen = fromUnitChild.split(sep)
            sep = '/' if '/' in toUnit else '*'
            if len(toUnit.split(sep)) == 2:
                toNum, toDen = toUnit.split(sep)
            else:
                raise NotImplementedError("conversion from single to triple or more not implmented.")
            numerator = _get_conversion(1, fromNum, toNum, getPath=False)
            denominator = 1 / _get_conversion(1, fromDen, toDen, getPath=False)
            if numerator is not None and denominator is not None:
                conversionFactor = baseConversion * numerator / denominator
                unitsNetwork.Memory[(fromUnit, toUnit)] = lambda x : x * conversionFactor
                return value * conversionFactor if value is not None else unitsNetwork.Memory[(fromUnit, toUnit)]

    # no conversion found
    raise NoConversionFoundError("from '" + str(fromUnit) + "' to '" + str(toUnit) + "'")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:38:58 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.1'
__release__ = 20220921
__all__ = ['unitProduct','unitDivision']

from .dictionaries import dictionary
from .convert import convertible
from .helpers.is_number import is_number
from .helpers.multisplit import multisplit


def unitSplit(unit_string):
    us = multisplit(unit_string)
    return us


def unitBasePower(unit_string):
    uBas, uPow = '', ''
    oth = ''
    inv = False
    if '/' in unit_string and len(unit_string.split('/')) == 2 and is_number(unit_string.split('/')[0]):
        invPow, unit_string = unit_string.split('/')
        inv = True

    for c in unit_string :
        if c.isdigit() :
            uPow += oth + c
            oth = ''
        elif c in ['-','+','.']:
            oth += c
        else :
            uBas += oth + c
            oth = ''
    uPow = 1 if uPow == '' else float(uPow) if '.' in uPow else int(uPow)
    if inv :
        uPow = uPow * -1 * (float(invPow) if '.' in invPow else int(invPow))
    return uBas, uPow


def unitBase(unit_string):
    return unitBasePower(unit_string)[0]


def unitProduct(unit_string1,unit_string2):
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in dictionary['dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in dictionary['dimensionless']:
        if unit_string2.lower().strip(' ()') not in dictionary['dimensionless']:
            return unit_string2
        else :
            return unit_string1

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unitProduct(unit_string1, unit_string1)

    U1bas, U1pow = unitBasePower(unit_string1)
    U2bas, U2pow = unitBasePower(unit_string2)

    if convertible(U1bas, U2bas):
        Upow = U1pow+U2pow
        if Upow == -1:
            result = U1bas+'-1'
        elif Upow == 1:
            result = U1bas
        elif Upow == 0:
            result = U1bas + '/' + U1bas
        else :
            for c in ['+','-','^']:  # '*','/'
                if c in U1bas:
                    U1bas = '('+U1bas+')'
                    break
            result = U1bas + str(Upow)
    else:
        for c in ['+','-','^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '('+unit_string1+')'
                break
        for c in ['+','-','^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '('+unit_string2+')'
                break
        result = unit_string1 + '*' + unit_string2
    return result


def unitDivision(unit_string1, unit_string2):
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in dictionary['dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in dictionary['dimensionless']:
        if unit_string2.lower().strip(' ()') not in dictionary['dimensionless']:
            uBas, uPow = unitBasePower('1/' + unit_string2)
            return uBas+str(uPow)
        else:
            return unit_string1

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unitDivision(unit_string1, unit_string1)

    U1bas, U1pow = unitBasePower(unit_string1)
    U2bas, U2pow = unitBasePower(unit_string2)

    if convertible(U1bas, U2bas):
        Upow = U1pow - U2pow
        if Upow == -1:
            result = U1bas + '-1'
        elif Upow == 1:
            result = U1bas
        elif Upow == 0:
            result = U1bas + '/' + U1bas
        else:
            for c in ['+','-','^']:  # '*','/'
                if c in U1bas:
                    U1bas = '('+U1bas+')'
                    break
            result = U1bas + str(Upow)

    elif ('+' not in unit_string1 and '-' not in unit_string1 and '^' not in unit_string1) and (
            '+' not in unit_string2 and '-' not in unit_string2 and '^' not in unit_string2) and (
                '*' in unit_string1 and unitBase(unit_string2) in map(unitBase, unit_string1.split('*'))):
            result = ''
            for u in unit_string1.split('*'):
                if unit_string2 == u:
                    return '*'.join([u for u in unit_string1.split('*') if u != unit_string2])
                elif unitBase(unit_string2) == unitBase(u):
                    result = (result + '*' + unitDivision(u, unit_string2).strip('*')).strip('*')
                else:
                    result = (result + '*' + u).strip('*')
    else:
        for c in ['+','-','^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '('+unit_string1+')'
                break
        for c in ['+','-','^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '('+unit_string2+')'
                break
        result = unit_string1 + '/' + unit_string2
    return result


def unitPower(unit_string, power):
    if unit_string is None:
        unit_string = 'dimensionless'
    if power is None:
        power = 1

    if type(unit_string) is str and len(unit_string.strip(' ()')) == 0:
        unit_string = 'dimensionless'
    if type(power) is str and len(power.strip(' ()')) == 0:
        power = 1

    U1bas, U1pow = unitBasePower(unit_string)

    if type(power) in (int, float):
        Upow = U1pow * power
        if Upow == 0:
            return 'dimensionless'
        elif Upow == 1:
            return U1bas
        else:
            return U1bas + str(Upow)

    elif type(power) is str:
        U2bas, U2pow = unitBasePower(power)
        Upow = U1pow * U2pow
        if Upow == 0:
            return 'dimensionless'
        elif Upow == 1:
            return U1bas + '^' + U2bas
        else:
            return U1bas + '^' + str(Upow) + U2bas

    else:
        raise TypeError('power must be units string or numeric')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['length', 'area', 'volume']

from ..dictionaries import dictionary
from ..unit_class import unit


class length(unit):
    classUnits = dictionary['length']
    def __init__(self, value, units):
        self.name = 'length'
        self.kind = length
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class area(unit):
    classUnits = dictionary['area']
    def __init__(self, value, units):
        self.name = 'area'
        self.kind = area
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class volume(unit):
    classUnits = dictionary['volume']
    def __init__(self, value, units):
        self.name = 'volume'
        self.kind = volume
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['rate', 'speed', 'velocity']

from ..dictionaries import dictionary
from ..unit_class import unit


class rate(unit):
    classUnits = dictionary['rate']
    def __init__(self, value, units):
        self.name = 'rate'
        self.kind = rate
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class speed(unit):
    classUnits = dictionary['speed']
    def __init__(self, value, units):
        self.name = 'speed'
        self.kind = speed
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


def velocity(value, units):
    return speed(value, units)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['pressure', 'weight', 'compressibility']

from ..dictionaries import dictionary
from ..unit_class import unit


class pressure(unit):
    classUnits = dictionary['pressure']
    def __init__(self, value, units):
        self.name = 'pressure'
        self.kind = pressure
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class weight(unit):
    classUnits = dictionary['weight']
    def __init__(self, value, units):
        self.name = 'weight'
        self.kind = weight
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class compressibility(unit):
    classUnits = dictionary['compressibility']
    def __init__(self, value, units):
        self.name = 'compressibility'
        self.kind = compressibility
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 12:14:51 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.1'
__release__ = 20220924
__all__ = ['dictionary','SI','OGF','DATA','StandardAirDensity','StandadEarthGravity']


StandardAirDensity = 1.225 # Kg/m3 or g/cc
StandadEarthGravity = 9.80665 # m/s2 or 980.665 cm/s2 from


# the dictionary that contains all the units definitions
dictionary = {}


# Sistema Internacional
SI = {
    'Y' : (lambda X: X*1E+24,lambda X: X*1E+48,lambda X: X*1E+72) , # yotta
    'Z' : (lambda X: X*1E+21,lambda X: X*1E+42,lambda X: X*1E+63) , # zetta
    'E' : (lambda X: X*1E+18,lambda X: X*1E+36,lambda X: X*1E+54) , # exa
    'P' : (lambda X: X*1E+15,lambda X: X*1E+30,lambda X: X*1E+46) , # peta
    'T' : (lambda X: X*1E+12,lambda X: X*1E+24,lambda X: X*1E+36) , # tera
    'G' : (lambda X: X*1E+09,lambda X: X*1E+18,lambda X: X*1E+27) , # giga
    'M' : (lambda X: X*1E+06,lambda X: X*1E+12,lambda X: X*1E+18) , # mega
    'K' : (lambda X: X*1E+03,)*3 , # with uppercase K is commonly used to express x1000
    'k' : (lambda X: X*1E+03,lambda X: X*1E+06,lambda X: X*1E+09) , # kilo
    'h' : (lambda X: X*1E+02,lambda X: X*1E+04,lambda X: X*1E+06) , # hecto
    'd' : (lambda X: X*1E-01,lambda X: X*1E-02,lambda X: X*1E-03) , # deci
    'c' : (lambda X: X*1E-02,lambda X: X*1E-04,lambda X: X*1E-06) , # centi
    'm' : (lambda X: X*1E-03,lambda X: X*1E-06,lambda X: X*1E-09) , # mili
    'µ' : (lambda X: X*1E-06,lambda X: X*1E-12,lambda X: X*1E-18) , # micro
    'u' : (lambda X: X*1E-06,lambda X: X*1E-12,lambda X: X*1E-18) , # micro
    'n' : (lambda X: X*1E-09,lambda X: X*1E-18,lambda X: X*1E-27) , # nano
    'p' : (lambda X: X*1E-12,lambda X: X*1E-24,lambda X: X*1E-36) , # pico
    'f' : (lambda X: X*1E-15,lambda X: X*1E-30,lambda X: X*1E-45) , # femto
    'a' : (lambda X: X*1E-18,lambda X: X*1E-36,lambda X: X*1E-54) , # atto
    'z' : (lambda X: X*1E-21,lambda X: X*1E-42,lambda X: X*1E-63) , # zepto
    'y' : (lambda X: X*1E-24,lambda X: X*1E-48,lambda X: X*1E-72) , # yocto
    }

SI_order = (('length','pressure','weight','mass','time',),('area',),('rate','volume',),)

DATA = {
    'Y' : (lambda X: X*1E+24,lambda X: X*2**80) , # yotta
    'Z' : (lambda X: X*1E+21,lambda X: X*2**70) , # zetta
    'E' : (lambda X: X*1E+18,lambda X: X*2**60) , # exa
    'P' : (lambda X: X*1E+15,lambda X: X*2**50) , # peta
    'T' : (lambda X: X*1E+12,lambda X: X*2**40) , # tera
    'G' : (lambda X: X*1E+09,lambda X: X*2**30) , # giga
    'M' : (lambda X: X*1E+06,lambda X: X*2**20) , # mega
    'K' : (lambda X: X*1E+03,lambda X: X*2**10) , # kilo with uppercase K because it is very common
    'k' : (lambda X: X*1E+03,lambda X: X*2**10) , # kilo
    }

DATA_order = (('dataBIT',),('dataBYTE',))


# Oil & Gas Field Unit System
OGF = { 'M' : (None,None,lambda X: X*1E+03),
        'MM' : (None,None,lambda X: X*1E+06),
        'B' : (None,None,lambda X: X*1E+09),
        'T' : (None,None,lambda X: X*1E+12),
    }
OGF_order = (tuple(),tuple,('volume','rate',))


dictionary['time'] = []
dictionary['time_NAMES_UPPER_PLURALwS_REVERSE'] = {
    'nanosecond': ('ns',),
    'millisecond': ('ms',),
    'second': ('s', 'sec',),
    'minute': ('min',),
    'hour': ('h', 'hr'),
    'day': ('d', 'día', 'días', 'DíA'),
    'week': ('w', 'we'),
    'month': ('mo', 'mes', 'meses'),
    'year': ('y', 'año'),
    'lustrum': tuple(),
    'decade': tuple(),
    'century': ('centuries',),
    }
dictionary['time_SI'] = ('s',)


# temperature and related
dictionary['temperature'] = ['C', 'K', 'F', 'R']
dictionary['temperature_NAMES'] = {
    'Celsius': ('Centigrades', 'C', 'DEG C', 'DEGREES C'),
    'Fahrenheit': ('F','DEG F','DEGREES F'),
    'Rankine': ('R','DEG R','DEGREES R'),
    'Kelvin': ('K','DEG K','DEGREES K')
    }
dictionary['temperatureGradient'] = []


# volume
dictionary['volume'] = []
dictionary['volume_SI_UPPER_REVERSE'] = ('m3', 'sm3', 'stm3', 'rm3',)  # 'l' # litre is volumen but the conversion of SI prefixes is linear
dictionary['volume_UK_NAMES_UPPER_REVERSE_PLURALwS'] = {
    'fuild ounce': ('fl oz','oz','ounce', 'ozUS', 'ounce'),
    'gill': ('gi', 'gillUS', 'giUS'),
    'pint': ('pt', 'pintUS', 'ptUS'),
    'quart': ('qt', 'quartUS', 'qtUS'),
    'gallonUS': ('gal', 'galUS', 'USgal', 'USgallon', 'gallon'),
    'gallonUK': ('imperial gallon', 'galUK', 'UKgal', 'UKgallon'),  # 'gal', 'gallon'
    'fuild ounce UK': ('fl oz UK','ozUK','ounceUK'),
    'gillUK': ('giUK',),
    'pintUK': ('ptUK',),
    'quartUK': ('qtUK',),
    }
dictionary['volume_NAMES_UPPER_REVERSE_PLURALwS_SPACES'] = {
    'litre': ('l', 'liter', 'litro'),
    'mililitre': ('ml', 'mililiter', 'cubic centimeter'),
    'centilitre': ('cl', 'centiliter'),
    'decilitre': ('dl', 'deciliter') ,
    'cubic meter': ('CM', 'm3'),
    'standard cubic meter': ('scm', 'sm3', 'stm3', 'm3'),
    'cubic centimeter': ('cc', 'cm3', 'standard cubic centimeter'),
    'standard cubic centimeter': ('scc', 'scm3'),
    'reservoir cubic meter': ('rm3',),
    'reservoir cubic centimeter': ('rcc', 'rcm3'),
    'cubic foot': ('cubic feet', 'ft3', 'cf', 'pie cúbico', 'pie cubico', 'pc'),
    'standard cubic foot': ('scf', 'cf'),
    'cubic inch': ('in3', 'cubic inches'),
    'barrel': ('bbl', 'stb'),
    'reservoir barrel': ('rb',),
    'standard barrel': ('stb', 'stbo', 'stbw', 'stbl', 'oil barrel'),
    }
dictionary['volume_UPPER_REVERSE'] =  ('kstm3', 'Mstm3')
dictionary['volume_PLURALwS'] = ('liter', 'mililiter', 'centiliter', 'deciliter', 'barrel', 'oil barrel', 'gals', 'UKgallons', 'USgallons', 'oil gallon')
dictionary['volume_OGF'] = ('scf', 'cf', 'ft3', 'stb', 'bbl', 'rb', 'stbo', 'stbw', 'stbl')
# dictionary['volume_oilgas_NAMES'] = ('scf','cf','ft3','stb','bbl','rb','stbo','stbw','stbl')
dictionary['volume_oilgas_UPPER'] =  ('sm3', 'm3', 'rm3', 'ksm3', 'Msm3', 'Gsm3',
                                      'scf', 'cf', 'ft3', 'Mscf', 'MMscf', 'Bscf', 'Tscf', 'Mcf', 'MMcf', 'Bcf', 'Tcf',
                                      'stb', 'bbl', 'rb', 'Mstb', 'MMstb', 'Bstb', 'Tstb', 'Mbbl', 'MMbbl', 'Mrb', 'MMrb')
dictionary['volume_product_NAMES'] = {
    'm3': ('m2*m',),
    'cm3': ('cm2*cm',),
    'ft3': ('ft2*ft',),
    'in3':('in2*in',)
    }


# length
dictionary['length'] = []
dictionary['length_NAMES_UPPER_REVERSE'] = {'meter': ('m', 'meter', 'metro')}
dictionary['length_SI'] = ('m', 'l')  # litre is volumen but the conversion of SI prefixes is linear
dictionary['length_UK_NAMES_UPPER_REVERSE'] = {
    'thou': ('th',),
    'inch': ('in', '"'),
    'foot': ('feet','ft',"'"),
    'yard': ('yd',),
    'chain': ('ch',),
    'rod': ('rd',),
    'furlong': ('fur',),
    'mile': ('mi',),
    'league': ('lea',),
    'nautical mile': ('nmi',),
    'nautical league': ('nlea',),
    }


# area
dictionary['area'] = []
dictionary['area_NAMES_UPPER_REVERSE'] = {'square meter': ('sq m', 'm2', 'sqmeter', 'm*m', 'm3/m')}
dictionary['area_SI'] = ('m2',)
dictionary['area_UK_NAMES_UPPER_REVERSE'] = {
    'square mile': ('sq mi','mi2','sqmile','mi*mi'),
    'acre': tuple(),
    'square rod': ('sq rd', 'sqrd', 'rd2', 'rd*rd'),
    'square yard': ('sq yd', 'sqyd', 'yd2', 'yd*yd'),
    'square foot': ('sq ft', 'sqft', 'ft2', 'ft*ft', 'ft3/ft'),
    'square inch': ('sq in', 'sqin', 'in2', 'in*in', 'in3/in')
    }


# pressure
dictionary['pressure'] = []
dictionary['pressure_NAMES_UPPER_REVERSE_SPACES'] = {
    'absolute psi': ('psia', 'lb/in2', 'absolute pound/square inch', 'psi absolute', 'libras/pulgada cuadrada absoluta', 'lpca'),
    'absolute bar': ('bara', 'barsa', 'abs bar', 'bar absolute'),
    'atmosphere': ('atm', 'atma'),
    'Pascal': ('Pa', 'Newton/m2'),
    'kPa': ('KPa', 'kilopascal'),
    'hPa': ('hectopascal',),
    'Torr': ('millimeters of mercury',),
    'millimeters of mercury': ('mmHg',),
    }
dictionary['pressure_NAMES_UPPER_SPACES'] = {
    'psi gauge': ('psi', 'pound/square inch', 'psig', 'gauge psi'),
    'bar gauge': ('bar', 'barg', 'gauge bar', 'bars'),
    }
dictionary['pressure_SI'] = ('Pa', 'bara', 'bar')

dictionary['pressureGradient'] = []
dictionary['pressureGradient'] = ('psi/ft', 'psia/ft', 'psig/ft', 'psi/m', 'psia/m', 'psig/m', 'bar/m', 'bars/m', 'barsa/m', 'bara/m', 'barg/m')


# weight
dictionary['weight'] = []
dictionary['weight_NAMES_UPPER_REVERSE_SPACES_PLURALwS'] = {
    'gram': ('g',),
    'kilogram': ('kg', 'kgm', 'Kgm', 'kilogram mass'),
    'milligrams': ('mg',),
    'metric ton': ('Tonne',),
    'g-mol': ('g-moles',),
    'Kg-mol': ('Kg-moles',),
    }
dictionary['weight_UK_NAMES_UPPER_REVERSE_SPACES_PLURALwS'] = {
    'grain': ('gr',),
    'pennyweight': ('pwt', 'dwt'),
    'dram': ('dr', 'dramch'),
    'ounce': ('oz',),
    'pound': ('lb', '#', 'libra'),
    'stone': ('st',),
    'quarter': ('qr', 'qrt'),
    # 'hundredweight' : ('cwt',),
    'short hundredweight': ('US hundredweight', 'UScwt'),
    'long hundredweight': ('UK hundredweight', 'UKcwt', 'cwt'),
    # 'ton' : ('t',),
    'short ton': ('USton',),
    'long ton': ('t', 'UKton', 'ton'),
    }
dictionary['weight_SI'] = ('g', 'g-mol')


# mass
dictionary['mass'] = ['kilogram mass']


# density
dictionary['density'] = []
dictionary['density_oilgas'] = {}
dictionary['density_NAMES_UPPER'] = {
    'API': ('DEGREES',),
    'SgG': ('gas-gravity','gas-specific-gravity'),
    'SgW': ('water-gravity',),
    'SgO': ('oil-gravity',),
    }
dictionary['density_NAMES_UPPER_REVERSE'] = {
    'g/cm3': ('g/cc',),
    'kg/m3': ('Kg/m3',),
    'lb/ft3': tuple(),
    'psi/ft': tuple(),
    'kJ/rm3': ('KJ/rm3',),
    'lb/stb': tuple(),
    'psia/ft': ('psi/ft',),
    'bara/m': ('bar/m',),
    }


# compressibility
dictionary['compressibility'] = []
dictionary['compressibility_UPPER_NAMES'] = {
    '1/psi': ('1/psia', 'µsip', 'usip', '1/psig'),
    'µsip': ('usip',),
    '1/bar': ('1/bara', '1/barg')
    }


# rate
dictionary['rate'] = []
dictionary['rate_NAMES_UPPER_SPACES_REVERSE'] = {
    'standard barrel per day': ('stb/day',),
    'standard cubic foot per day': ('scf/day', 'cf/day', 'scfd'),
    'standard cubic meter per day': ('sm3/day',),
    'barrel per day': ('bbl/day'),
    'cubic meter per day': ('m3/day',),
    'cubic foot per day': ('ft3/day',),
    'reservoir barrel per day': ('rb/day',),
    'reservoir cubic meter per day': ('rm3/day',),
    }
dictionary['rate_NAMES_UPPER_SPACES_REVERSE'] = {
    'stb/day': ('stbd',),
    'scf/day': ('scfd', 'cf/day',),
    'sm3/day': ('sm3d', 'stm3d', 'stm3/day'),
    'bbl/day': ('bbld',),
    'm3/day': ('m3/d',),
    'ft3/day': ('cf/day',),
    }


# digital data
dictionary['dataBYTE'] = []
dictionary['dataBYTE_UPPER_PLURALwS_DATA_NAME_REVERSE'] = {'byte': ('Byte',)}
dictionary['dataBYTE_DATA_REVERSE'] = ('B',)
dictionary['dataBIT'] = []
dictionary['dataBIT_DATA_PLURALwS_NAME_REVERSE'] = {'bit': ('Bit', 'BIT',)}
dictionary['dataBIT_DATA_REVERSE'] = ('b',)


# viscosity
dictionary['viscosity'] = []
dictionary['viscosity_UPPER_NAMES_REVERSE'] = {
    'centipoise': ('cP',),
    'Poise': ('dyne*s/cm2', 'g/cm/s'),
    'Pa*s': ('N*s/m2', 'kg/m/s')
    }


# permeability
dictionary['permeability'] = []
dictionary['permeability_UPPER_REVERSE'] = ('mD', 'Darcy',)


# force
dictionary['force'] = []
dictionary['force_NAMES_SPACES_RECURSIVE_UPPER_REVERSE'] = {
    'Newton': ('N', 'newton', 'kg*m/s2'),
    'kilogram force': ('kgf', 'kilopondio',),  # 'kilogram'
    'kilopondio': ('kp',),
    'Dyne': ('dyne', 'dyn', 'g*cm/s2')
    }


 # energy
dictionary['energy'] = []
dictionary['energy_UPPER_NAMES_REVERSE'] = {
    'Joule': ('J', 'Watt second', 'N*m', 'kg*m2/s2', 'Joules'),
    'Kilojoule': ('kJ',),
    'kilowatt hour': ('kWh', 'kW*h', 'kilovatio hora'),
    'British thermal unit': ('BTU',),
    'Watt second': ('Watt*second', 'W*s', 'Ws'),
    'Watt hour': ('Wh', 'W*h',),
    'gram calorie': ('cal', 'calorie', 'Calorie'),
    }
dictionary['energy_SI'] = ('Wh', 'Ws',)


# power
dictionary['power'] = []
dictionary['power_UPPER_NAMES_REVERSE'] = {
    'Horsepower': ('hp',),
    'Watt': ('W', 'J/s'),
    }
dictionary['power_SI'] = ('W',)


# productivity index
dictionary['productivityIndex'] = []
dictionary['productivityIndex_UPPER_NAMES_REVERSE'] = {
    'stb/day/psi': ('STB/DAY/', 'stbd/psi', 'stbd/psia', 'stb/day/psia', 'stb/day-psi', 'STB/DAY-PSI', 'stb/day-psia', 'STB/DAY-PSIA','stb/d/psi'),
    'sm3/day/bar': ('SM3/DAY/', 'sm3/d/b', 'sm3d/bar', 'sm3d/bara', 'sm3/day/bara', 'sm3/day-bar', 'SM3/DAY-BAR','sm3/day-bara', 'SM3/DAY-BARA', 'sm3/day/barsa'),
    'sm3/day/kPa': ('sm3d/kPa', 'sm3d/kPa', 'sm3/day-kPa', 'SM3/DAY-KPA', 'sm3/d/kPa')
    }


# acceleration
dictionary['acceleration'] = ('m/s2', 'ft/s2',)


# dimensionless
dictionary['dimensionless'] = []
dictionary['dimensionless_fractions_UPPER_NAMES'] = {'fraction': ('ratio', 'dimensionless', 'unitless', 'None', '')}

dictionary['percentage'] = []
dictionary['percentage_NAMES_REVERSE'] = {'percentage': ('%', 'perc'),}


# dates
dictionary['date'] = []
dictionary['date_UPPER_PLURALwS'] = ['date']


dictionary['userUnits'] = []


# other
dictionary['otherUnits'] = []
dictionary['otherUnits_UPPER_NAMES'] = {
    'sec/day': ('sec/d',),
    's2': ('s*s',)
    }


temperatureRatioFactors = {'Celsius':9,
                           'Fahrenheit':5,
                           'Kelvin':9,
                           'Rankine':5,
                           }
temperatureRatioConversions = {}
for t1, c1 in temperatureRatioFactors.items():
    for t1a in ((t1,) + dictionary['temperature_NAMES'][t1]):
        for t2, c2 in temperatureRatioFactors.items():
            for t2a in ((t2,) + dictionary['temperature_NAMES'][t2]):
                temperatureRatioConversions[(t1a,t2a)] = c1/c2
del temperatureRatioFactors


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 13:12:52 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>

helper functions for units modules
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['multisplit']


def multisplit(string, sep=['*','/'], remove=[' ']):
    """
    receives a string and returns a list with string split by all the separators in sep.
    the default separator is the blank space ' '.
    use the remove parameter to indicate the separators that must not be reported in the output list.
    by default, the blank space is not reported.
    """
    assert type(string) is str

    # check sep is list
    if type(sep) is str:
        sep = [sep]

    # eliminate duplicated separators
    sep = list(set(sep))

    # sort sep by lenght
    s = len(sep)
    for i in range(s-1) :
        for j in range(s-i-1) :
            if len(sep[j]) < len(sep[j+1]):
                sep[j], sep[j+1] = sep[j+1], sep[j]

    # initialize counters
    stringlist = []
    i, x, t = 0, 0, len(string)
    # loop through the entire string
    while i < t:
        found = False # flag for found separator
        # look for each separator
        for se in sep :
            s = len(se)
            if (i+s <= t) and string[i:i+s] == se:
                stringlist += [string[x:i], se]
                x = i+s
                i += s
                found = True
                break
        i += 1 if not found else 0
    stringlist += [string[x:]]

    # clean the output
    newlist = []
    for part in stringlist :
        if part not in remove + [''] :
            newlist += [part]

    return tuple(newlist)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['density', 'volumeRatio', 'productivityIndex', 'pressureGradient']

from ..dictionaries import dictionary
from ..unit_class import unit


class density(unit):
    classUnits = dictionary['density']
    def __init__(self, value, units):
        self.name = 'density'
        self.kind = density
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class volumeRatio(unit):
    classUnits = dictionary['volumeRatio']
    def __init__(self, value, units):
        self.name = 'volumeRatio'
        self.kid = volumeRatio
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class productivityIndex(unit):
    classUnits = dictionary['productivityIndex']
    def __init__(self, value, units) :
        self.name = 'productivityIndex'
        self.kind = productivityIndex
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)

class pressureGradient(unit):
    classUnits = dictionary['pressureGradient']
    def __init__(self, value, units) :
        self.name = 'pressureGradient'
        self.kind = pressureGradient
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:46:04 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['is_unit']

from ..dictionaries import dictionary


def is_unit(unit):
    if type(unit) is str:
        unit = unit.strip()

    for each in list(dictionary.keys()):
        if '_' not in each:
            isU = unit in list(dictionary[each])
            if isU :
                return True

    if '/' in unit or '*' in unit:
        unitSplit = []
        for each in unit.split('/'):
            unitSplit += each.split('*')
        ret = [False] * len(unitSplit)

        for each in list(dictionary.keys()):
            if '_' not in each :
                for subU in range(len(unitSplit)):
                    if unitSplit[subU] in list(dictionary[each]):
                        ret[subU] = True

        for each in ret:
            if not each:
                return False

    return True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 21:17:35 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['energy', 'power']

from ..dictionaries import dictionary
from ..unit_class import unit


class energy(unit):
    classUnits = dictionary['energy']
    def __init__(self, value, units):
        self.name = 'energy'
        self.kind = energy
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


class power(unit):
    classUnits = dictionary['power']
    def __init__(self, value, units):
        self.name = 'power'
        self.kind = power
        self.value = self.checkValue(value)
        self.unit = self.checkUnit(units)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 18:24:20 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.4'
__release__ = 20220924
__all__ = ['units', 'convert']


from .units.define import units
from .convert import convertUnit
from .convert import convertUnit as convert
from .unit_class import unit

def print_path(switch=None):
    from ._database import unitsNetwork
    if type(switch) is str and switch.lower().strip() == 'off':
        switch = False
    elif type(switch) is str:
        unitsNetwork.print = bool(switch)
    else:
        unitsNetwork.print = False if unitsNetwork.print is True else True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['unit']

from .errors import WrongUnitsError, WrongValueError, NoConversionFoundError
from .operations import unitProduct, unitDivision, unitBasePower
from .convert import converter, convertible

import numpy as np
from pandas import Series, DataFrame
from numbers import Number

arraylike = [np.ndarray]
try:
    from pandas import Series, DataFrame
    arraylike += [Series, DataFrame]
except:
    pass
arraylike = tuple(arraylike)


class unit(object):
    """
    A class to cope with values associated to units, its arithmetic and logic
    operations and conversions.
    """

    def __init__(self, value=None, unit=None):
        self.unit = None
        self.value = None
        self.name = None
        self.kind = None
        if value is not None and unit is not None:
            from .units.define import units
            u = units(value, unit)
            self.unit = u.unit
            self.value = u.value
            self.name = u.name
            self.kind = u.kind

    def __call__(self):
        return self.value

    def __repr__(self):
        return str(self.value) + '_' + str(self.unit)

    def __str__(self) :
        return str(self.value) + '_' + str(self.unit)

    def convert(self, newunit):
        if type(newunit) is not str:
            try:
                newunit = newunit.unit
            except:
                raise WrongUnitsError("'" + str(newunit) + "' for '" + str(self.name) + "'")
        return self.kind(converter(self.value, self.unit, newunit), newunit)

    def to(self, newunit):
        return self.convert(newunit)

    def __neg__(self):
        return self.kind(self.value * -1, self.unit)

    def __bool__(self):
        from .units.custom import userUnits
        from .units.unitless import dimensionless, percentage
        if self.kind in (dimensionless, percentage, userUnits):
            return False
        return True

    def __abs__(self):
        return self.kind(abs(self.value), self.unit)

    def __add__(self, other):
        from .units.unitless import dimensionless, percentage
        if '.units.' in str(type(other)):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value + other.value, self.unit)
                elif convertible(other.unit, self.unit):
                    return self.kind(self.value + converter(other.value, other.unit, self.unit), self.unit)
                elif convertible(self.unit, other.unit):
                    return self.kind(other.value + converter(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return self.kind(self.value + other.value, other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value + other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value + other.value, self.unit)
            else:  # add different units
                return (self, other)
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self + each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is percentage:
            return self.kind((self.value + other) * 100,self.unit)
        else:
            return self.kind(self.value + other, self.unit)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return other.kind(self.value * other.value, other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value * other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value * other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value * other.value, unitProduct(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value * converter(other.value, other.unit, self.unit), unitProduct(self.unit, other.unit))
            elif convertible(self.unit, other.unit):
                return units(other.value * converter(self.value, self.unit, other.unit), unitProduct(other.unit, self.unit))
            elif convertible(unitBasePower(self.unit)[0], unitBasePower(other.unit)[0]):
                factor = converter(1, unitBasePower(other.unit)[0], unitBasePower(self.unit)[0])
                return units(self.value * other.value * factor, unitProduct(self.unit, other.unit))
            else:
                return units(self.value * other.value, unitProduct(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self * each]
            return tuple(result)
        elif self.kind is percentage:
            return self.kind(self.value * other * 100, self.unit)
        else:
            return self.kind(self.value * other, self.unit)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return other.kind(self.value ** other.value, other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value ** other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value ** other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
            elif convertible(other.unit, self.unit):
                return units(self.value ** converter(other.value, other.unit, self.unit), self.unit + '^' + self.unit)
            elif convertible(unitBasePower(self.unit)[0], unitBasePower(other.unit)[0]):
                factor = converter(1, unitBasePower(other.unit)[0], unitBasePower(self.unit)[0])
                return units(self.value ** (other.value * factor), self.unit + '^' + other.unit)
            else:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self ** each]
            return tuple(result)
        else:
            if self.kind is percentage:
                return self.kind((self.value ** other) * 100, self.unit)
            else:
                powunits = unitBasePower(self.unit)[1] * other
                if powunits == 0:
                    powunits = 'dimensionless'
                elif powunits == 1:
                    powunits = unitBasePower(self.unit)[0]
                else:
                    powunits = unitBasePower(self.unit)[0] + str(powunits)
                return units(self.value ** other, powunits)

    def __rpow__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if self.kind is percentage:
            return units((other ** self.value), 'dimensionless')
        elif self.kind is dimensionless:
            return units(other ** self.value, self.unit)
        else:
            raise TypeError("unsupported operand type(s) for ** or pow(): '" + str(type(other)) + "' and '" + self.name)

    def __sub__(self, other):
        # return self.__add__(other * -1)
        from .units.unitless import dimensionless, percentage
        if '.units.' in str(type(other)):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value - other.value, self.unit)
                elif convertible(other.unit, self.unit):
                    return self.kind(self.value - converter(other.value, other.unit, self.unit), self.unit)
                elif convertible(self.unit, other.unit):
                    return self.kind(other.value - converter(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return self.kind(self.value - other.value, other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value - other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value - other.value, self.unit)
            else:  # add different units
                return (self, other)
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self - each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is percentage:
            return self.kind((self.value - other) * 100, self.unit)
        else:
            return self.kind(self.value - other, self.unit)

    def __rsub__(self, other):
        return -self + other

    def __truediv__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return other.kind(self.value / other.value, '1/' + other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value / other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value / other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value / other.value, unitDivision(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value / converter(other.value, other.unit, self.unit), unitDivision(self.unit, other.unit))
            elif convertible(unitBasePower(self.unit)[0], unitBasePower(other.unit)[0]):
                factor = converter(1, unitBasePower(other.unit)[0], unitBasePower(self.unit)[0])
                return units(self.value / (other.value * factor), unitDivision(self.unit, other.unit))
            else:
                return units(self.value / other.value, unitDivision(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self / each]
            return tuple(result)
        elif self.kind is percentage:
            return self.kind(self.value / other * 100, self.unit)
        else:
            return units(self.value / other, self.unit)

    def __rtruediv__(self, other):
        from .units.define import units
        from .units.unitless import dimensionless, percentage
        if self.kind is percentage:
            return units(other / self.value * 100, self.unit)
        elif self.kind is dimensionless:
            return units(other / self.value, self.unit)
        else:
            return units(other / self.value, self.unit + '-1')

    def __floordiv__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return other.kind(self.value // other.value, '1/' + other.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value // other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value // other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value // other.value, unitDivision(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value // converter(other.value, other.unit, self.unit), unitDivision(self.unit, other.unit))
            elif convertible(unitBasePower(self.unit)[0], unitBasePower(other.unit)[0]):
                factor = converter(1, unitBasePower(other.unit)[0], unitBasePower(self.unit)[0])
                return units(self.value // (other.value * factor), unitDivision(self.unit, other.unit))
            else:
                return units(self.value // other.value, unitDivision(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self // each]
            return tuple(result)
        elif self.kind is percentage:
            return self.kind(self.value // other * 100, self.unit)
        else:
            return self.kind(self.value // other, self.unit)

    def __rfloordiv__(self, other):
        from .units.define import units
        from .units.unitless import dimensionless, percentage
        if self.kind is percentage:
            return units(other // self.value * 100, self.unit)
        elif self.kind is dimensionless:
            return units(other // self.value, self.unit)
        else:
            return units(other // self.value, self.unit + '-1')

    def __mod__(self, other):
        from .units.unitless import dimensionless, percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (dimensionless, percentage) and other.kind not in (dimensionless, percentage):
                return other.kind(self.value % other.value, self.unit)
            elif self.kind is percentage and other.kind in (dimensionless, percentage):
                return self.kind((self.value % other.value) * 100, self.unit)
            elif other.kind in (dimensionless, percentage):
                return self.kind(self.value % other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value % other.value, self.unit)
            elif convertible(other.unit, self.unit):
                return units(self.value % converter(other.value, other.unit, self.unit), self.unit)
            elif convertible(unitBasePower(self.unit)[0], unitBasePower(other.unit)[0]):
                factor = converter(1, unitBasePower(other.unit)[0], unitBasePower(self.unit)[0])
                return units(self.value % (other.value * factor), self.unit)
            else:
                return units(self.value % other.value, self.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self % each]
            return tuple(result)
        elif self.kind is percentage:
            return self.kind(self.value % other * 100, self.unit)
        else:
            return self.kind(self.value % other, self.unit)

    def __lt__(self, other):
        if type(self) == type(other):
            return self.value < other.convert(self.unit).value
        else:
            msg = "'<' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __le__(self, other):
        if type(self) == type(other):
            return self.value <= other.convert(self.unit).value
        else :
            msg = "'<=' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.','') + "'"
            raise TypeError(msg)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.value == other.convert(self.unit).value
        else:
            msg = "'==' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.','') + "'"
            raise TypeError(msg)

    def __ne__(self, other):
        if type(self) == type(other):
            return self.value != other.convert(self.unit).value
        else:
            msg = "'!=' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.','') + "'"
            raise TypeError(msg)

    def __ge__(self, other):
        if type(self) == type(other):
            return self.value >= other.convert(self.unit).value
        else :
            msg = "'>=' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.','') + "'"
            raise TypeError(msg)
    def __gt__(self, other):
        if type(self) == type(other):
            return self.value > other.convert(self.unit).value
        else :
            msg = "'>' not supported between instances of '" + \
                (str(type(self))[str(type(self)).index("'")+1:len(str(type(self))) - str(type(self))[::-1].index("'")-1]).replace('__main__.','') + \
                    "' and '" + \
                        (str(type(other))[str(type(other)).index("'")+1:len(str(type(other))) - str(type(other))[::-1].index("'")-1]).replace('__main__.','') + "'"
            raise TypeError(msg)

    def __len__(self):
        try:
            return len(self.value)
        except:
            return 1

    def __getitem__(self, item):
        if type(item) is int:
            if item >= len(self):
                raise IndexError
        else:
            raise ValueError
        from .units.unitless import percentage
        if self.kind is percentage:
            return self.kind(self.value[item] * 100, self.unit)
        else:
            return self.kind(self.value[item], self.unit)

    def __iter__(self):
        if type(self.value) in (int, float):
            return np.array((self.value,)).__iter__()
        else:
            return self.value.__iter__()

    def getUnit(self):
        return self.unit

    def getValue(self):
        return self.value

    def checkValue(self, value):
        if type(value) in (list, tuple):
            try:
                return np.array(value)
            except:
                raise WrongValueError(str(value))
        elif isinstance(value, Number):
            return value
        elif isinstance(value, arraylike):
             return value
        else:
            raise WrongValueError(str(value))

    def checkUnit(self, units):
        if type(units) is not str:
            try:
                units = units.unit
            except:
                raise WrongUnitsError("'" + str(units) + "' for '" + str(self.name) + "'")
        if units in self.kind.classUnits:
            return units
        else:
            raise WrongUnitsError("'" + str(units) + "' for '" + str(self.name) + "'")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.0'
__release__ = 20220920
__all__ = ['units']


from ..dictionaries import dictionary
from .custom import userUnits, otherUnits
from .force import pressure, weight, compressibility
from .geometry import length, area, volume
from .temperature import temperature, temperatureGradient
from .time import time
from .unitless import dimensionless, percentage
from .ratios import density, volumeRatio, productivityIndex, pressureGradient
from .rates import rate, speed, velocity
from .energy import energy, power

from numpy import ndarray
from numbers import Number

arraylike = [ndarray]
try:
    from pandas import Series, DataFrame
    arraylike += [Series, DataFrame]
except:
    pass
arraylike = tuple(arraylike)


def units(value, units=None):
    """
    return an instance of units with the provided value and units.

    Parameters
    ----------
    value : int, float, numeric array, Series or DataFrame
        the value to assign the units.
    units : str, optional
        the units to assign the value. The default is None.

    Raises
    ------
    TypeError
        if value or units are not the appropriate kind.

    Returns
    -------
    units
        instance of units.

    """
    if units is None and '.units.' in str(type(value)):
        value, units = value.value, value.units
    if units is None:
        units = 'dimensionless'
    if type(units) is not str:
        raise TypeError("'units' must be a string.")
    units = units.strip()
    for kind in dictionary:
        if units in dictionary[kind]:
            if isinstance(value, Number):
                return eval(kind + "(" + str(value) + ",'" + units + "')")
            elif isinstance(value, arraylike):
                u = eval(kind + "(" + str(0) + ",'" + units + "')")
                u.value = value
                return u
            else:
                raise TypeError("'value' parameter must be numeric.")

    return userUnits(value, units)